# parse smalt -a -f sam output
# Substitute bases from the reference to which read aligns
# produce new Fastq output for read
VERBOSE = True
DEBUG = True

ALI_GAP_CHAR = "-"
WINDOW_HALFSIZ = 6

class AliBlock:
    BLOCK_START_KEY = "QUERY:"
    BLOCK_END_KEY = "REFERENCE:"
    def __init__(self):
        self._blank()
        
    def _blank(self):
        self.status = 0
        self.lin = "" # last line read
        self.qs = int(0)
        self.qseq = ""
        self.qe = int(0)
        self.alistr = ""
        self.ss = int(0)
        self.sseq = ""
        self.se = int(0)
        self.ok = False
        self.refcheck = False
        self.linctr = 0
        
    def check(self):
        return len(self.qseq) == len(self.sseq)
    
    def parse(self, infil):
        termflg = False
        self.linctr = 0
        while not termflg:
            self.lin = infil.readline()
            if not self.lin:
                termflg = True
                self.okflg = False
                break
            self.linctr = self.linctr + 1
            fld = self.lin.split()
            if len(fld) > 0:
                break
        if termflg:
            return termflg
        #print "PARSE:", fld
        if fld[0] != AliBlock.BLOCK_START_KEY:
            termflg = True
            self.ok = False
            return termflg
        #print "Key Ok."
        if len(fld) < 4:
            print "Format error", fld
            self.ok = False
            return termflg
        self.qs = int(fld[1])
        self.qseq = fld[2]
        self.qe = int(fld[3])
        self.alistr = infil.readline()
        self.lin = infil.readline()
        self.linctr = self.linctr + 1
        fld = self.lin.split()
        if len(fld) != 4 or fld[0] != AliBlock.BLOCK_END_KEY:
            #print "End not ok"
            self.ok = False
            return termflg
        #print fld
        self.linctr = self.linctr + 1
        self.ss = int(fld[1])
        self.sseq = fld[2]
        self.se = int(fld[3])
        self.ok = True
        #print "End ok"

        return termflg

    def add(self, other, with_refcheck=True):
        isReverse = self.qe < self.qs
        if isReverse and other.qe > other.qs:
            return False
        if isReverse:
            if other.qe > other.qs:
                return False
            if other.qs + 1 != self.qe:
                self.ok = False
        else:
           if other.qs > other.qe: 
               return False
           if other.qs != self.qe + 1:
               self.ok = False
        if not self.ok:
            if isReverse:
                sense = "REVERSE"
            else:
                sense = "FOWARD"
            print "Query ends [%s] don't agree %i %i" % (sense, other.qs, self.qe)
        
        if with_refcheck and self.ok and other.ss != self.se + 1:
            print "Reference ends don't agree %i %i" % (other.ss, self.se)
            self.ok = False

        if self.ok:
            self.qe = other.qe
            self.se = other.se
            self.qseq += other.qseq
            self.alistr += other.alistr
            self.sseq += other.sseq
        return self.ok
    
class Alignment:
    BLOCK = AliBlock()
    def __init__(self):
        self.lin = ""
        self.ablk = AliBlock()
        self.ok = False

    def parse(self, infil, is_verbose=False, with_refcheck=True):
        termflg = self.ablk.parse(infil)
        if is_verbose:
            print "Parsed 1st block: %s" % self.ablk.qseq
        self.ok = self.ablk.ok
        self.lin = self.ablk.lin
        isReverse = self.ablk.se < self.ablk.ss
        ctr = 0
        while self.ok and not termflg:
            ctr = ctr + 1
            termflg = Alignment.BLOCK.parse(infil)
            if is_verbose:
                print "parsed block %i, termflg=%s, ok=%s" % (ctr,termflg,Alignment.BLOCK.ok)
            if termflg:
                self.lin = Alignment.BLOCK.lin
                break
            if Alignment.BLOCK.ok:
                self.ablk.add(Alignment.BLOCK, with_refcheck)
                if not self.ablk.ok:
                    self.ok = False
        return termflg

    def checkNrun(self, max_start=10, fromEnd=False, verbose=False):
        # cut alignemnt starting with up to 10 bases followed by > 20 Ns in reference
        # if fromEnd == True, check alignment ending with 20Ns followed by up to 10 bases
        ssq = self.ablk.sseq
        alilen = len(self.ablk.qseq)
        isN = False
        lastN = -1
        numN = 0
        ri = range(alilen)
        if fromEnd:
            ri.reverse()
        j = 0
        for i in ri:
            if verbose:
                print "[%i] '%s'" % (i, ssq[i])
            if ssq[i] == 'N':
                lastN = j
                if isN:
                    numN += 1
                else:
                    numN = 1
                    isN = True
                if verbose:
                    print "[%i] '%s' (N) lastN = %i, numN = %i" % (i, ssq[i], lastN, numN)
            elif j < max_start:
                if verbose and isN:
                    print "[%i] '%s' (start)" % (i, ssq[i])
                isN = False
            else:
                break
            j += 1

        if lastN < 0:
            nonNpos = 0
        else:
            nonNpos = lastN + 1
        return (nonNpos, numN)
    
    def fillN(self, verbose=False):
        # subsitute string of 'N's in query with reference bases
        qstr = ""
        qsq = self.ablk.qseq
        if not 'N' in qsq:
            return ["", True, 0]
        ssq = self.ablk.sseq
        alilen = len(self.ablk.qseq)
        q_added = 0
        is_win_clear = True
        for i in range(alilen):
            q = qsq[i]
            s = ssq[i]
            if q == 'N':
                if s == ALI_GAP_CHAR:
                    if is_win_clear:
                        if verbose:
                            print "Keep '%s:%s' at pos %i" % (s,q,i)
                        qstr = qstr + q
                    else:
                        if verbose:
                            print "Skip '%s:%s' at pos %i" % (s,q,i)
                        q_added = q_added - 1
                else:
                    if verbose:
                        print "substitute '%s:%s' at pos %i" % (s,q,i)
                    qstr = qstr + s
                    is_win_clear = False
            elif q == ALI_GAP_CHAR:
                if is_win_clear:
                    last_qgap_pos = i
                if not is_win_clear and s != ALI_GAP_CHAR:
                    print "add '%s:%s' at pos %i" % (s,q,i)
                    qstr = qstr + s
                    q_added = q_added + 1
            else:
                qstr = qstr + q
                if not is_win_clear and q == s:
                    if verbose:
                        print "Switch to clear window at pos %i '%s:%s'" % (i,s,q)
                    is_win_clear = True
 
               
        if self.ablk.qe > self.ablk.qs:
            targetlen = self.ablk.qe - self.ablk.qs + 1
        else:
            targetlen = self.ablk.qs - self.ablk.qe + 1
        okflg = (targetlen + q_added) == len(qstr)
        if not okflg:
            print "Error substituted sequence length %i differs from target %i (added %i)" % \
                  (len(qstr), targetlen, q_added)
        return [qstr, okflg, q_added]

    def subst(self, windowhalfsiz, verbose=False):
        qstr = ""
        qsq = self.ablk.qseq
        ssq = self.ablk.sseq
        alilen = len(self.ablk.qseq)
        is_win_clear = False
        for i in range(alilen):
            q = qsq[i]
            s = ssq[i]
            if q == s:
                qstr = qstr + q
                if i > (last_gap_pos + windowhalfsiz) or last_gap_pos < 1:
                    if verbose and not is_win_clear:
                        print "Switch to clear window at pos %i '%s:%s'" % (i,q,s)
                    is_win_clear = True
            elif q == ALI_GAP_CHAR:
                if verbose and is_win_clear:
                    print "Toggle to blocked window at pos %i '%s:%s'" % (i,q,s)
                last_gap_pos = i
                is_win_clear = False
            elif s == ALI_GAP_CHAR:
                if verbose and is_win_clear:
                    print "Toggle to blocked window at pos %i '%s:%s'" % (i,q,s)
                qstr = qstr + q
                last_gap_pos = i
                is_win_clear = False
            elif is_win_clear and s != 'N':
                j_max = i + windowhalfsiz
                if j_max > alilen: j_max = alilen
                for j in range(i+1, j_max):
                    if qsq[j] == ALI_GAP_CHAR or ssq[j] == ALI_GAP_CHAR:
                        if verbose:
                            print "Toggle to blocked while looking ahead at pos %i from %i '%s:%s'" \
                              % (j,i,q,s)
                        is_win_clear = False
                        last_gap_pos = j
                if is_win_clear:
                    qstr = qstr + s
                    if verbose:
                        print "mismatch substitute '%s:%s' at pos %i" % (q,s,i)
                else:
                    qstr = qstr + q
                    if verbose:
                        print "mismatch keep '%s:%s' at pos %i" % (q,s,i)
            else:
               qstr = qstr + q
               if verbose:
                   print "mismatch keep '%s:%s' at pos %i" % (q,s,i)
               
        if self.ablk.qe > self.ablk.qs:
            targetlen = self.ablk.qe - self.ablk.qs + 1
        else:
            targetlen = self.ablk.qs - self.ablk.qe + 1
        okflg = targetlen == len(qstr)
        if not okflg:
            print "Error substituted sequence length %i differs from target %i" % \
                  (len(qstr), targetlen)
        return [qstr, okflg]

def writeAsFasta(sam, ali, qstr, n_added, oufil):
    if len(qstr) > 1:
        blk = ali.ablk
        isReverse = blk.qe < blk.qs
        qlen = len(sam.nt)
        if isReverse:
            qs = qlen - blk.qs
            qe = qlen - blk.qe
        else:
            qs = blk.qs - 1
            qe = blk.qe - 1

        if qs < 0:
            exit("ERROR: writeAsFastq() negative coordinates")

        if len(qstr) != (qe - qs + 1) + n_added:
            exit("ERROR in coordinate calculation qs=%i (%i), qe=%i (%i), qlen=%i (%i), rev=%s" % \
                 (qs, blk.qs, qe, blk.qe, qlen, len(qstr), isReverse))

        if qs > 0:
            fullstr = sam.nt[0:qs] + qstr
        else:
            fullstr = qstr

        if qe + 1 < qlen:
            fullstr += sam.nt[qe+1:qlen]

        if len(sam.nt) + n_added != len(fullstr):
            #print sam.nt
            #print fullstr
            #print qstr
            #print qs, qe, blk.qs, blk.qe, qlen
            exit("ERROR in subsitution %i (target %i)" % (len(fullstr), len(sam.nt)))
        sam.nt = fullstr
    namstr, seq, qual = sam.asFastqStr()
    oufil.write(">%s\n%s\n" % (namstr, seq))
    
def processSubstitutions(infil, oufil, is_verbose):
    from SAM import Sam
    from sys import exit
    sam = Sam()
    ali = Alignment()
    termflg = False
    lin = infil.readline()
    readctr = 0
    while 1:
        if not lin:
            break
        sam.parse(lin, is_verbose)
        if not sam.ok:
            lin = infil.readline()
            continue
        if is_verbose or DEBUG:
            print "\n** read: %s" % (sam.qname)

        termflg = ali.parse(infil, is_verbose)
        if ali.ok:
            [qseq, ok,n_added] = ali.fillN(DEBUG)
            if not ok:
                exit("Error substituted sequence length differs from target")
            if is_verbose:
                print "[%s] %s" % (ok, qseq)
            writeAsFasta(sam, ali, qseq, n_added,oufil)          
            readctr += 1
            termflg = False
            lin = ali.lin
            if is_verbose:
                print "NEWLINE (%s): %s" % (termflg, lin)
                
    return readctr
        
if __name__ == '__main__':
    from sys import argv, exit
    from SAM import openFile
    
    if len(argv) != 3:
        exit("usage: %s <smalt_output> <FASTQ output>" % argv[0])

    infilnam = argv[1]
    oufilnam = argv[2]
    #winhalfsiz = int(argv[2])
    #if winhalfsiz < 1: winhalfsiz = WINDOW_HALFSIZ
    infil = openFile(infilnam, 'r')
    oufil = openFile(oufilnam, 'w')
    nreads = processSubstitutions(infil, oufil, VERBOSE)
    print "Number of reads processed: %i" % (nreads)
    print "Fasta file: %s" % (oufilnam)
    oufil.close()
    infil.close()

    exit()
