## from mapping output in SSAHA format, check whether
## the reference start with TTAA signature (might not be
## part of the mapping, and may contain 1 mismatch).

class Ssaha:
    from re import compile
    DEBUG = True
    FORMLIN = compile("alignment:([ABCDNSR]):(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+" + \
                      "(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+([FC*])\s+(\d+)\s+([\.\d]+)\s+(\d+)\s+(\d+)")
    def __init__(self):
        self._blank()
        
    def _blank(self):
        self.mapcls = ''
        self.mapq = 0
        self.swscor = 0
        self.qnam = ''
        self.rnam = ''
        self.qseg = ()
        self.sseg = ()
        self.sense = '' # direction 'F' or 'C'
        self.alilen = 0
        self.id = float(0)
        self.rlen = 0
        self.slen = 0
        self.ok = False
    
    def parse(self, lin):
        m = Ssaha.FORMLIN.match(lin)
        if m:
            self.mapcls = m.group(1)    # mapping class label
            self.mapq = int(m.group(2)) # mapping quality
            self.swscor = int(m.group(3)) # Smith-Waterman score
            self.qnam = m.group(4)      # read name
            self.rnam = m.group(5)      # name of reference
            self.qseg = (int(m.group(6)), int(m.group(7)))
            self.sseg = (int(m.group(8)), int(m.group(9)))
            self.sense = m.group(10)
            self.alilen = int(m.group(11))
            self.id = float(m.group(12))
            self.rlen = int(m.group(13))
            self.slen = int(m.group(14))
            self.ok = True
        else:
            if Ssaha.DEBUG: print "not parsed: %s" % (lin)
            self.ok = False

    def __str__(self):
        s = "alignment:%s:%2.2i %i" % (self.mapcls, self.mapq, self.swscor)
        s = s + " %s %s" % (self.qnam, self.rnam)
        s = s + " %i %i" % tuple(self.qseg)
        s = s + " %i %i" % tuple(self.sseg)
        s = s + " %s %i %5.2f %i %i" % (self.sense, self.alilen, self.id, self.rlen, self.slen)
        return s
    
def openFile(filnam, mode):
    import gzip
    from sys import exit

    is_compressed = len(filnam) > 3 and filnam[-3:] == ".gz"
    try:
        if is_compressed:
            oufil = gzip.open(filnam, mode)
        else:
            oufil = open(filnam, mode)
    except:
        print "ERROR when opening file '%s'" % filnam
        exit(1)

    return oufil        

if __name__ == '__main__':
    from sys import argv, exit

    if len(argv) < 2:
        print "usage: %s <ssaha file>" % (argv[0])
        exit(1)

    infilnam = argv[1]

    ssa = Ssaha()
    
    infil = openFile(infilnam, 'r')

    while 1:
        lin = infil.readline()
        if not lin: break

        ssa.parse(lin)
        if ssa.ok:
            print "%s %s" % (ssa.qnam, ssa.ok)
